# this little script reads hex dump files from the Sol
# BASIC/5 program and detokenizes the binary.
# you can tell it is from BASIC/5 because the starting
# address of the dump is 1AD9.
#
# there are a few things we must be aware of:
#    1) keywords are tokenized; all have values 0x80 or greater
#    2) Line format:
#          aa bb cc blah blah blah<CR>
#          aa bb cc blah blah blah<CR>
#       where blah blah blah is a mix of tokens and straight ascii.
#       'aa' is the length of the line, in bytes.
#       bb+256*cc is the line number.
#
# Other info:
#   BASIC/5  lines start at 1AD9
#   EXTBASIC lines start at 3C20

# token mapping

# this one is from BASIC5
%tokens = (
	    0x80, 'LET ',
	    0x81, 'NEXT ',
	    0x82, 'IF ',
	    0x83, 'GOTO ',
	    0x84, 'GOSUB ',
	    0x85, 'RETURN',
	    0x86, 'READ ',
	    0x87, 'DATA ',
	    0x88, 'FOR ',
	    0x89, 'PRINT ',
	    0x8A, 'INPUT ',
	    0x8B, 'DIM ',
	    0x8C, 'STOP',
	    0x8D, 'END',
	    0x8E, 'RESTORE ',
	    0x8F, 'REM ',
	    0x90, 'CLEAR',
	    0x91, 'SET ',
	    0x92, 'FILE ',
	    0x93, 'CLOSE ',
	    0x94, 'BYE',
	    0x95, ': ',
	    0x96, ';',
	    0x9C, 'TAB',
	    0x9D, ' THEN ',
	    0x9E, ' TO ',
	    0x9F, ' STEP ',
	    0xA0, 'RUN',
	    0xA1, 'LIST',
	    0xA2, 'NEW',
	    0xA3, 'SAVE',
	    0xA4, 'GET',
	    0xA5, 'EDIT',
	    0xA6, 'XEQ',
	    0xC4, 'SQR',
	    0xC6, 'INT',
	    0xCC, 'ARG',
	    0xCD, 'CALL',
	    0xCE, 'RND',
	    0xD2, 'SGN',
	    0xD3, 'SIN',
	    0xD7, 'TAN',
	    0xD8, 'COS',
	    0xE0, '(',
	    0xE2, '*',
	    0xE3, '+',
	    0xE5, '-',
	    0xE7, '/',
	    0xEF, '>=',
	    0xF0, '<=',
	    0xF1, '<>',
	    0xF4, '<',
	    0xF5, '=',
	    0xF6, '>',
          );


# get file name
$fname = shift;

# open file
open(INFILE, $fname) || die "Couldn't open '$fname', $!";

#slurp in the file as a list of words, dropping <CR>s
while(<INFILE>) {
    chomp;
    if (/^EN/) { next }
    s/[0-9A-F]{4}:\s*//;
    push @file, split;
}

$START = 1;
$MID   = 2;

$state = $START;

while(@file) {

    if ($state == $START) {
	if ($#file < 2) { die "expecting more bytes" };
	$byte1 = hex nextbyte();	# length of line, in bytes
	$byte2 = hex nextbyte();
	$byte3 = hex nextbyte();
	$lineno = $byte3*256 + $byte2;
	printf "%-4d ", $lineno;
    }
    $state = $MID;

    $byte = nextbyte();
    $val = hex $byte;

    if (($val >= 32) and ($val < 127)) {

	# simple ascii character
	print chr($val);

    } elsif ($val == 127) {

	# ascii DEL character, which is unprintable
	print "<DEL>";

    } elsif ($val > 127) {

	# it is a token; look it up and print it if a mapping exists
	$tok = $tokens{$val};
	if ($tok) {
	    if ($tok eq "#LINE#") {
		# dummy token; it marks that the next two bytes are
		# a line number, so we must interpret them
		if ($#file < 1) { die "expecting more bytes" };
		$byte1 = hex nextbyte();
		$byte2 = hex nextbyte();
		$lineno = $byte2*256 + $byte1;
		print $lineno;
	    } else {
		print $tok;
	    }
	} else {
	    print "<$byte>";
	}

    } elsif ($val == 13) {

	# end of line
	print "\n";
	$state = $START;

    } else {
	print "<$byte>";
    }
}

print "\n";

close(INFILE);


# grab the next token from the input string, stripping out
# the address bytes
sub nextbyte {

    my($b,$len);

    while (@file) {

	$b = shift @file;
	$len = length($b);

	if ($len == 2) {
	    return $b;
	}

	if ($len != 4) { die "got token '$b'"; }

	# print "<$b>";	# print 4-digit hex address, for debugging
    }

    return "<EOF>";
}

